function [choice probabilities] = eSoft(values,explorationRate)

    % eSoft solutions do the best action 1-e percent of the time. e percent
    % of the time they make a uniform choice
    
    % determine the number of choices
    numberOfChoices = length(values);
    
    % assign a position value to each choice
    possibleChoices = 1:1:numberOfChoices;
    
    % determine the best choice and control for ties
    [maxValue choice] = max(values);
    [tiesTrue tieLocations] = find(values == maxValue);
    if sum(tiesTrue) > 1
        % reasign max location to one of the tie values
        choice = possibleChoices(randi(length(tiesTrue)));
    end

    % check to see if we go with the max choice or if we explore    
    doIChooseMax = rand(1);
    if doIChooseMax < 1 - explorationRate
    else
        choice = randi(numberOfChoices);
    end
    
    % determine output probabilites
    probabilities(1:1:length(values)) = explorationRate / numberOfChoices-1;
    probabilities(choice) = 1 - explorationRate;
    
end